/*-
 * Copyright (C)2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: cmd.c,v 1.29 2021/07/16 22:39:12 babolo Exp $"

#define BLIN_COMPAT     4
#define MIFE_COMPAT     5
#define Bpars_COMPAT    4
#define RECOBE_COMPAT   4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <zlib.h>
#include <babolo/BLINflag.h>
#include <babolo/recobe.h>
#include <babolo/parser.h>
#include <mife.h>
#include "pngtr.h"
#include "headers.h"

static void
/*****************************************************************************************************
 ****                                                               ****
 *****************************************************************************************************
                                                                                                  ****/
bsp(int o, const char c[], u_int64_t s) {                                                        /****
                                                                                                  ****
 *****************************************************************************************************/
    if  (!o) {
        ifBLIN_QX0("%s %02X %02X %02X %02X %02X %02X %02X %02X"
                  , c
                  , (s >> 56) & 0x0FF
                  , (s >> 48) & 0x0FF
                  , (s >> 40) & 0x0FF
                  , (s >> 32) & 0x0FF
                  , (s >> 24) & 0x0FF
                  , (s >> 16) & 0x0FF
                  , (s >> 8) & 0x0FF
                  , s & 0x0FF
                  );
    } else {
        ifBLIN_QX1("%s %02X %02X %02X %02X %02X %02X %02X %02X"
                  , c
                  , (s >> 56) & 0x0FF
                  , (s >> 48) & 0x0FF
                  , (s >> 40) & 0x0FF
                  , (s >> 32) & 0x0FF
                  , (s >> 24) & 0x0FF
                  , (s >> 16) & 0x0FF
                  , (s >> 8) & 0x0FF
                  , s & 0x0FF
                  );
}   }

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     num-                                                 ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_realconvey(pngtr_conf *cf) {                                                               /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    u_int8_t         *wholepng;
    pngtr_convey     *convey;
    u_int             chctr;     /* chunk counter */
    pngtr_hdr        *hdr;       /* one chunk header */
    int               ex    = EX_OK;
    ssize_t           p;         /* wholepng moving pointer */

    ifBLIN_QX3("+ %08X %u", cf->flags, cf->num);
    PNGTR_X((cf->num >= PNGTR_CONVEYNUM),            ("Ill convey num %d", cf->num), ERANGE, EX_USAGE);
    convey = &cf->convey[cf->num];
    PNGTR_X(!(convey->png.cont),                             ("Convey not ready"), ENOENT, EX_NOINPUT);
    wholepng = convey->png.cont;
    p = 8;
    if  (PNGTR_NOSIGN & convey->flags) {
        p = 0;
    } else {
        u_int64_t sgn  = recobe_load(wholepng, -8);

        convey->IHDR.type = 0;
        for (int i = 1; i <= PNGTR_SIGNA_NUM; ++i) {
            if  (sgn == pngtr_sigs[i]) {
                convey->IHDR.type = i;
                break;
        }   }
        if  (!convey->IHDR.type) {
            bsp((PNGTR_IGNSIGN & convey->flags), "Bad signature", sgn);
            bsp((PNGTR_IGNSIGN & convey->flags), "Good for PNG:", pngtr_sigs[1]);
            bsp((PNGTR_IGNSIGN & convey->flags), "Good for MNG:", pngtr_sigs[2]);
            bsp((PNGTR_IGNSIGN & convey->flags), "Good for JNG:", pngtr_sigs[3]);
            if  (PNGTR_IGNSIGN & ~convey->flags) {
                errno = EFTYPE;
                ex = EX_DATAERR;
                goto out;
    }   }   }
    ifBLIN_QX4( "%08X %"BLIN_X" %u %u"
              , convey->flags
              , BLIN_I(wholepng)
              , (u_int)p
              , (u_int)convey->png.size
              );
    for (chctr = 0; p < convey->png.size; ++chctr) {
        hdr = (void *)&wholepng[p];
        p += 12 + recobe_load(hdr->ln, -4);
        ifBLIN_QX4( "%u %"BLIN_X" %"BLIN_X" %u %u"
                  , chctr
                  , BLIN_I(wholepng)
                  , BLIN_I(hdr)
                  , (u_int)recobe_load(hdr->ln, -4)
                  , (u_int)p
                  );
    }
    convey->chunks = calloc(chctr + 1, sizeof(pngtr_chunk));
    PNGTR_W(!(convey->chunks),                                       ("chunks x %u", chctr), EX_OSERR);
    convey->chumax = convey->chunum = chctr;
    p = (PNGTR_NOSIGN & convey->flags) ? 0 : 8;
    ifBLIN_QX4("%"BLIN_X" %u %u", BLIN_I(wholepng), (u_int)p, (u_int)convey->png.size);
    for (chctr = 0; chctr < convey->chunum; ++chctr) {
        u_int32_t         chunklen;
        pngtr_chunk      *chunk;
        ssize_t           q;         /* hdr moving pointer */

        chunk = &convey->chunks[chctr];
        hdr = (void *)&wholepng[p];
        chunklen = 12 + recobe_load(hdr->ln, -4);
        ifBLIN_QX4("%u %"BLIN_X" %08X %u", chctr, BLIN_I(hdr), (u_int)p, chunklen);
        chunk->offset = p;
        if  (((PNGTR_AGRERR & convey->flags) ? 8 : 12) <= chunklen) {
            const pngtr_chdef  *chdef;
            const pngtr_chidef *opdef;
            u_int               opnum;

        ;   chunk->type = babolo_testword(&headers, hdr->nm);
        ;   if  (  (25 <= chunklen)
                && (PNGTR_PNGHEAD & pngtr_chdescr[chunk->type].shift)
                ) {                          /* chunklen, sizeof(pngtr_IHDR) + 12 == 25 rounds to 28 */
        ;       switch(convey->IHDR.type) {
                case PNGTR_TYPE_PNG:
        ;           bcopy(&hdr->n, &convey->IHDR, 13);
        ;           break;
                case PNGTR_TYPE_MNG:
        ;           bcopy(&hdr->n, &convey->IHDR, 8);
        ;           break;
                case PNGTR_TYPE_JNG:
        ;           bcopy(&hdr->n, &convey->IHDR, 8);
        ;           convey->IHDR.depth = ((pngtr_JHDR *)&hdr->n)->adepth;
        ;           convey->IHDR.colortype = ((pngtr_JHDR *)&hdr->n)->colortype;
        ;           convey->IHDR.compress = ((pngtr_JHDR *)&hdr->n)->acompress;
        ;           convey->IHDR.filter = ((pngtr_JHDR *)&hdr->n)->afilter;
        ;           convey->IHDR.interlace = ((pngtr_JHDR *)&hdr->n)->ainterlace;
                default:
        ;           ifBLIN_QX0("Unkn type %u", convey->IHDR.type);
        ;           errno = ENOSYS;
        ;           ex = EX_DATAERR;
        ;           goto out;
        ;       }
        ;       convey->IHDR.width = recobe_load((void *)&convey->IHDR.width, -4);
        ;       convey->IHDR.height = recobe_load((void *)&convey->IHDR.height, -4);
        ;       convey->flags |= PNGTR_SELECTD;
        ;   }
        ;   chdef = &pngtr_chdescr[chunk->type];
        ;   opdef = chdef->ch[(PNGTR_CHCOLOR & chdef->shift) ? convey->IHDR.colortype : 0];
        ;   PNGTR_W(!(chunk->operands = calloc(opdef->opn, sizeof(pngtr_operand))),    (""), EX_OSERR);
        ;   q = p + 8;
        ;   opnum = opdef->opn;
        ;   for (u_int opctr = 0; opctr < opnum; ++opctr) {
                u_int32_t        oplen;

        ;       if  ((ssize_t)(p + chunklen) > convey->png.size) chunklen = convey->png.size - p;
        ;       ifBLIN_QX4("%"BLIN_X" %"BLIN_X" %"BLIN_X, p, p + chunklen, q);
        ;       if  ((ssize_t)(p + chunklen) <= q) break;
        ;       oplen = p + chunklen - q;
        ;       if  (!(PNGTR_AGRERR & convey->flags)) {
        ;           if  (4 >= oplen) break;
        ;           oplen -= 4;
        ;       }
        ;       ifBLIN_QX4("%u o %08X %u", opctr, q, oplen);
        ;       if  ((PNGTR_CHDEP & opdef->op[opctr].var) && (wholepng[q] != 8)) opdef = chdef->ch[1];
        ;       if  ((PNGTR_CHZIP & opdef->op[opctr].var) && !!wholepng[q]) opdef = chdef->ch[1];
        ;       if  (PNGTR_OPNUM & opdef->op[opctr].var) opnum = 6 + wholepng[q];
        ;       chunk->carg = opctr + 1;
        ;       chunk->operands[opctr].val = (void *)&wholepng[q];
        ;       chunk->operands[opctr].size = (opdef->op[opctr].minl < oplen)
                                            ? opdef->op[opctr].minl
                                            : oplen
        ;       ;
        ;       chunk->operands[opctr].maxindex = 1;
        ;       if  (PNGTR_OPFIX & opdef->op[opctr].var) {
        ;           chunk->operands[opctr].maxindex = opdef->op[opctr].maxl;
        ;       }
        ;       if  (PNGTR_OPIDX & opdef->op[opctr].var) {
        ;           chunk->operands[opctr].maxindex = oplen / opdef->op[opctr].minl;
        ;       }
        ;       if  (PNGTR_OPVAR & opdef->op[opctr].var) {
        ;           chunk->operands[opctr].size = strnlen((char *)&wholepng[q], oplen);
        ;           if  ((0 < oplen) && !wholepng[q + chunk->operands[opctr].size]) ++q;
        ;       }
        ;       ifBLIN_QX4("%u o %08X", opctr, chunk->operands[opctr].size);
        ;       q += chunk->operands[opctr].maxindex * chunk->operands[opctr].size;
        }   }
        if  ((ssize_t)(p + chunklen) > convey->png.size) {
            chunk->csz = convey->png.size - p;
            p = convey->png.size;
        } else {
            chunk->csz = chunklen;
            p += chunklen;
    }   }
    PNGTR_O(!!(ex = pngtr_imgfla(cf)),     ("pngtr_ingfla %s", !convey->png.nm ? "" : convey->png.nm));
out:
    if  (!!ex) pngtr_freeconvey(cf);
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****    lastchunk  cf->num                                          ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_crechunk(pngtr_conf *cf, pngtr_chtype type) {                                              /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey   *convey;
    pngtr_chunk    *chunk;
    int             ex = EX_OK;

    ifBLIN_QX3("+ %08X q%u", cf->flags, cf->num);
    if  (!cf->convey || !(convey = &cf->convey[cf->num])) goto out;
    if  (!convey->chunks || (convey->chumax <= convey->chunum)) {
        convey->chunks = realloc(convey->chunks, (convey->chunum + 2) * sizeof(pngtr_chunk));
        PNGTR_W(!(convey->chunks),                      ("chunks x %u", convey->chunum + 2), EX_OSERR);
    }
    for (u_int chctr = convey->chunum; ; --chctr) {
        chunk = &convey->chunks[chctr];
        bcopy(chunk, &convey->chunks[chctr + 1], sizeof(pngtr_chunk));
        if  (chctr <= convey->lastchunk) break;
    }
    convey->chumax++;
    convey->chunum++;
    bzero(chunk, sizeof(pngtr_chunk));
    chunk->flags = PNGTR_VIRTUAL;
    chunk->type = type;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}
