/*-
 * Copyright (C)2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: sel.c,v 1.4 2021/11/04 21:46:08 babolo Exp $"

#define BLIN_COMPAT     4
#define MIFE_COMPAT     5
#define Bpars_COMPAT    4
#define RECOBE_COMPAT   4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <zlib.h>
#include <babolo/BLINflag.h>
#include <babolo/recobe.h>
#include <babolo/parser.h>
#include <mife.h>
#include "pngtr.h"
#include "headers.h"

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     min..max   num                                  ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_chunksel(pngtr_conf *cf, u_int min, u_int max) {                                           /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey = &cf->convey[cf->num];
    int               ex    = EX_OK;

    ifBLIN_QX3("+ %08X q%u c%u-%u", cf->flags, cf->num, min, max);
    convey->flags |= PNGTR_CHAINSL;
    for (u_int chctr = min; (chctr <= max) && (chctr < convey->chunum); ++chctr) {
        convey->chunks[chctr].flags |= PNGTR_SELECTD;
    }
    ifBLIN_QX4(" %"BLIN_X" %u", convey->chunks, convey->chunum);
    if  (max > convey->chunum) {
        convey->chunks = realloc(convey->chunks, (max + 1) * sizeof(pngtr_chunk));
        ifBLIN_QX4(" %"BLIN_X" %u", convey->chunks, max + 1);
        PNGTR_W(!convey->chunks,                                    ("realloc %u", cf->num), EX_OSERR);
        for (u_int chctr = convey->chunum + 1; chctr <= max; ++chctr) {
            bzero(&convey->chunks[chctr], sizeof(pngtr_chunk));
            if  (chctr >= min) convey->chunks[chctr].flags |= PNGTR_SELECTD;
        }
        convey->chumax = convey->chunum = max + 1;
    }
    convey->lastchunk = max;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     min..max  lastchunk    num             ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_opsel(pngtr_conf *cf, u_int min, u_int max) {                                              /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey = &cf->convey[cf->num];
    pngtr_chunk      *chunk  = &convey->chunks[convey->lastchunk];
    int               ex     = EX_OK;

    ifBLIN_QX3("+ %08X q%u c%u o%u-%u", cf->flags, cf->num, convey->lastchunk, min, max);
    PNGTR_X(!chunk,                                                   ("No chunk"), ENOATTR, EX_USAGE);
    convey->flags |= PNGTR_OPERSEL;
    for (u_int opctr = min; (opctr <= max) && (opctr < chunk->carg); ++opctr) {
        chunk->operands[opctr].flags |= PNGTR_SELECTD;
    }
    ifBLIN_QX4(" %"BLIN_X" %u", chunk->operands, chunk->carg);
    if  (max > chunk->carg) {
        chunk->operands = realloc(chunk->operands, (max + 1) * sizeof(pngtr_chunk));
        ifBLIN_QX4(" %"BLIN_X" %u", chunk->operands, max + 1);
        PNGTR_W(!chunk->operands,             ("realloc %u %u", cf->num, convey->lastchunk), EX_OSERR);
        for (u_int opctr = chunk->carg + 1; opctr <= max; ++opctr) {
            bzero(&chunk->operands[opctr], sizeof(pngtr_chunk));
            if  (opctr >= min) chunk->operands[opctr].flags |= PNGTR_SELECTD;
        }
        chunk->carg = max + 1;
    }
    convey->lastoperand = max;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     min..max  lastoperand  lastchunk           ****
 ****   num                                                                            ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_idxsel(pngtr_conf *cf, u_int min, u_int max) {                                             /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey = &cf->convey[cf->num];
    pngtr_chunk      *chunk  = &convey->chunks[convey->lastchunk];
    pngtr_operand    *op;
    int               ex     = EX_OK;

    ifBLIN_QX3( "+ %08X q%u c%u o%u x%u-%u"
              , cf->flags
              , cf->num
              , convey->lastchunk
              , convey->lastoperand
              , min
              , max
              );
    PNGTR_X(!chunk,                                                   ("No chunk"), ENOATTR, EX_USAGE);
    op = &chunk->operands[convey->lastoperand];
    convey->flags |= PNGTR_INDEXSL;
    if  (!op->select) {
        PNGTR_W( !(op->select = calloc(max, 1))
               , ( "calloc %s %u[%u %u] %u-%u"
                 , !convey->png.nm ? "" : convey->png.nm
                 , cf->num
                 , convey->lastchunk
                 , convey->lastoperand
                 , min
                 , max
                 )
               , EX_OSERR
               );
        op->maxindex = max;
    }
    for (u_int ixctr = min; (ixctr < max) && (ixctr < op->maxindex); ++ixctr) op->select[ixctr] |= 1;
    if  (max > op->maxindex) {
        PNGTR_W( !(op->select = realloc(op->select, max))
               , ( "realloc %s %u[%u %u] (%u)%u-%u"
                 , !convey->png.nm ? "" : convey->png.nm
                 , cf->num
                 , convey->lastchunk
                 , convey->lastoperand
                 , op->maxindex
                 , min
                 , max
                 )
               , EX_OSERR
               );
        for (u_int ixctr = op->maxindex; ixctr < max; ++ixctr) {
            op->select[ixctr] = (ixctr < min) ? 0 : 1;
        }
        op->maxindex = max;
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}
