/*-
 * Copyright (C)2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: pngtr.c,v 1.36 2021/11/04 21:46:08 babolo Exp $"

#define BLIN_COMPAT     4
#define MIFE_COMPAT     5
#define Bpars_COMPAT    4
#define RECOBE_COMPAT   4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <zlib.h>
#include <babolo/BLINflag.h>
#include <babolo/recobe.h>
#include <babolo/parser.h>
#include <mife.h>
#include "lib/pngtr.h"
#include "headers.h"

static const u_int8_t rotates[4] =
{0x0, 0x3, 0x6, 0x5} /* PNGTR_ROTATEY, PNGTR_ROTATEX, PNGTR_ROTSWAP */
;

static void
/*****************************************************************************************************
                                                                                                  ****/
usage() {                                                                                  /****
                                                                                                  ****
 *****************************************************************************************************/
    fprintf( stderr
           , PKG" @BABOLO V.M "VERS" http://www.babolo.ru/\n"
             "Usage: "PKG" [-options][ infile]\n"
             "   -a ANGLE      Rotate for -pp or -d\n"
             "   -b            Binary mode\n"
             "   -B            Text mode\n"
             "   -c N or N-N   Select chunks by number\n"
             "   -C            Unselect chunks by number\n"
             "   -d WHICH      diffenence curent image with another\n"
             "              N     Another conveyer number for difference\n"
             "              c     Total pixel count\n"
             "              n     File 1 name\n"
             "              N     File 2 name\n"
             "              p     Different pixel count\n"
             "              v     Pixels value normalized diffenence sum\n"
             "   -D            Delete selected chunks\n"
             "   -e            Extracts chunks to OUNM directory\n"
             "   -E -X XX      Use XX as env name for opt's arg\n"
             "   -f OUNM       Output file or directory name\n"
             "   -F FILE       Input binary image file\n"
             "   -G            Ignore signature error\n"
             "   -i N PATH     Prepare Nth conveyer with PNG file PATH\n"
             "   -I N          Switch to Nth conveyer\n"
             "   -L L          Forbid Lth interlace step\n"
             "   -m TYPE       Create chunk type TYPE\n"
             "   -o N or N-N   Select operands by number\n"
             "   -O            Clear all operand selections\n"
             "   -p WHICH      What to print, t text mode, b - binary mode\n"
             "              c     tb Chunk dump\n"
             "              e     tb Dump inflated\n"
             "              f     tb Image data with filters\n"
             "              F     t  Filters from image data\n"
             "              h     t  Chunk headers\n"
             "              i     tb Clean image data\n"
             "              n     t  File name\n"
             "              o     t  Chunk parameter\n"
             "              p      b PAM image\n"
             "              x     tb Dump big data\n"
             "   -q            Silent\n"
             "   -s            Proceed without signature in file begin\n"
             "   -S            Add signature to file begin\n"
             "   -t TYPE       Select chunks by type\n"
             "   -T            Unselect chunks by type\n"
             "   -v[v[v...]]   Verbose\n"
             "   -x N or N-N   Select indexes by number\n"
             "   -X            Unselect indexes by number\n"
             "   -y            No inflate validate\n"
             "   -Y            Aggressive short chunk data extraction\n"
             "   -z TEXT       Parce TEXT as arguments continuation\n"
             "   -Z TEXT       Parce TEXT as arguments continuation with env substitution\n"
             "   -= MODE       Internal diagnostic\n"
             "   -?            This text\n"
           )
    ;
}

int
/*****************************************************************************************************
                                                                                                  ****/
doclo(pngtr_conf *cf) {                                                                          /****
                                                                                                  ****
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    int               ex    = EX_OK;

    if  (!cf) goto out;
    ifBLIN_QX3("+ %08X %"BLIN_X, cf->flags, cf->file);
    if  (!!cf->file) {
        if  (PNGTR_BINOUT & cf->flags) {
            PNGTR_W(0 > close(fileno(cf->file)),            ("close %d", fileno(cf->file)),  EX_IOERR);
        } else {
            PNGTR_W(!!fclose(cf->file),                                        ("fclose"),   EX_IOERR);
        }
        cf->file = NULL;
        free(cf->ounm);
        cf->ounm = NULL;
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
                                                                                                  ****/
diapa(pngtr_conf *cf, babolo_opts *bos, pngtr_diapa *mimax) {                                    /****
                                                                                                  ****
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    int               ex    = EX_OK;
    char             *oo;
    const char       *o;

    PNGTR_W(!(o = babolo_getoptsarg(bos)),                           ("No -%c arg", bos->c), EX_USAGE);
    errno = 0;
    mimax->max = mimax->min = strtoul(o, &oo, 0);
    if  (errno) {
        ifBLIN_QW0("-%c strtoul=%s~", bos->c, o);
        ex = EX_USAGE;
        goto out;
    } else if (!!*oo && ('-' == *(oo++))) {
        errno = 0;
        mimax->max = strtoul(oo, &oo, 0);
        if  (errno) {
            ifBLIN_QW0("-%c strtoul=%s~", bos->c, o);
            ex = EX_USAGE;
            goto out;
    }   }
    if  (!!*oo) {
        ifBLIN_QX0("-%c text=%s~", bos->c, o);
        ex = EX_USAGE;
        errno = EPROTO;
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
                                                                                                  ****/
main(int argc, char **argv) {                                                                    /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
    const char       *flar  = "a:bBc:d:Def:F:Gi::I:L:m:o:Op:qsSt:Tvx:XyY=:?";
    const char       *Tflag = NULL;
    int               ex    = EX_OK;
    babolo_opts      *bos   = NULL;
    pngtr_conf       *cf    = NULL;
    const char       *o;
    int               c;
    int               i;

    PNGTR_W(!(cf = calloc(1, sizeof(pngtr_conf))),                                 ("conf"), EX_OSERR);
    blin_ctl(BLIN_CTL_FLAG | BLIN_CTL_FSET, (cf->flags = BLIN_BIT0));
#   define blin_internal_flags (cf->flags)
    cf->convey = calloc(PNGTR_CONVEYNUM, sizeof(pngtr_convey));
    PNGTR_W(!(cf->convey),                                                     ("convey_t"), EX_OSERR);
    PNGTR_W(!(bos = babolo_openopts(Bpars_ESYM | Bpars_SYME, 'Z')),         ("openopts"), EX_SOFTWARE);
    PNGTR_W(babolo_setopts(bos, Bpars_NONU, argc, argv, flar),               ("setopts"), EX_SOFTWARE);
    for (; !!(c = babolo_getopts(bos));) {
        ifBLIN_QX4("-%02X", c);
        PNGTR_W((0 > c),                                                        ("getopts"), EX_USAGE);
        switch (c) {
        case 'a': i = babolo_getoptunum(bos, 0);
                ; PNGTR_W((!!errno),                             ("getoptunum -%c", bos->c), EX_USAGE);
                ; PNGTR_W(!!(i % 90),                   ("angle %d is not factor of 90", i), EX_USAGE);
                ; cf->convey[cf->num].flags &= ~PNGTR_ROTMASK;
                ; i = (i / 90) % 4;
                ; if  (0 > i) i += 4;
                ; cf->convey[cf->num].flags |= rotates[i] << PNGTR_ROTSHFT;
                ; break;
        case 'b': PNGTR_O(!!(ex = doclo(cf)),                                               ("doclo"));
                ; cf->flags |= PNGTR_BINOUT;
                ; break;
        case 'B': PNGTR_O(!!(ex = doclo(cf)),                                               ("doclo"));
                ; cf->flags &= ~PNGTR_BINOUT;
                ; break;
        case 'c': {   pngtr_diapa mimax;

                ;     PNGTR_O(!!(ex = diapa(cf, bos, &mimax)),                ("Err -%c arg", bos->c));
                ;     PNGTR_O(!!(ex = pngtr_chunksel(cf, mimax.min, mimax.max)),         ("chunksel"));
                ; }
                ; break;
        case 'C': {   pngtr_convey   *convey = &cf->convey[cf->num];

                ;     convey->flags &= ~PNGTR_CHAINSL;
                ;     for (u_int chctr = 0; chctr < convey->chunum; ++chctr) {
                ;         convey->chunks[chctr].flags &= ~PNGTR_SELECTD;
                ; }   }
                ; break;
        case 'd': {   BLIN_flag f = 0;
                ;
                ;     i = 0;
                ;     PNGTR_W(!(o = babolo_getoptsarg(bos)),         ("No -%c arg", bos->c), EX_USAGE);
                ;     for (int j = 0; !!o[j]; ++j) {
                ;         switch(o[j]) {
                ;         case '0':
                ;         case '1':
                ;         case '2':
                ;         case '3':
                ;         case '4':
                ;         case '5':
                ;         case '6':
                ;         case '7':
                ;         case '8':
                ;         case '9': i = i * 10 + o[j] - '0'; break;
                ;         case 'c': f |= PNGTR_DIFFCOD;      break;
                ;         case 'n': f |= PNGTR_PRINTFN;      break;
                ;         case 'N': f |= PNGTR_PRINTF2;      break;
                ;         case 'p': f |= PNGTR_DIFFPNT;      break;
                ;         case 'v': f |= PNGTR_DIFFVAL;      break;
                ;         default : ifBLIN_QW0("Ill arg -%c %c", bos->c, o[j]);
                ;     }   }
                ;     PNGTR_X( (i >= PNGTR_CONVEYNUM)
                             , ("conveyer %d out of bonds", i)
                             , ENOENT
                             , EX_USAGE
                             );
                ;     if  (!cf->file) {
                ;         if  (!cf->ounm) {
                ;             cf->file = stdout;
                ;         } else {
                ;             cf->file = fopen(cf->ounm, "w");
                ;             PNGTR_W(!cf->file,                     ("fopen %s", cf->ounm), EX_IOERR);
                ;     }   }
                      PNGTR_O(!!(ex = pngtr_cmpam(cf, i, f, cf->file)),                     ("cmpam"));
                ; }
                ; break;
        case 'D': PNGTR_O(!!(ex = pngtr_chudelete(cf)),                                 ("chudelete"));
                ; break;
        case 'e': PNGTR_O(!!(ex = pngtr_dir(cf)),                                             ("dir"));
                ; break;
        case 'f': PNGTR_W(!(o = babolo_getoptsarg(bos)),             ("No -%c arg", bos->c), EX_USAGE);
                ; PNGTR_O(!!doclo(cf),                                                      ("doclo"));
                ; if  (!!cf->ounm) {
                ;     ifBLIN_QX1("Output name redefined from %s to %s", cf->ounm, o);
                ;     free(cf->ounm);
                ; }
                ; PNGTR_W(!(cf->ounm = strdup(o)),                               ("No mem"), EX_OSERR);
                ; break;
        case 'F': PNGTR_W(!(o = babolo_getoptsarg(bos)),             ("No -%c arg", bos->c), EX_USAGE);
                ; PNGTR_O(!!(ex = pngtr_read(cf, o, 1)),                       ("read 1 %u", cf->num));
                ; PNGTR_O(!!(ex = pngtr_realimg(cf)),                                     ("realimg"));
                ; break;
        case 'G': cf->flags |= PNGTR_IGNSIGN;
                ; break;
        case 'i': cf->num = babolo_getoptunum(bos, 0);
                ; PNGTR_W((!!errno),                             ("getoptunum -%c", bos->c), EX_USAGE);
                ; PNGTR_W(!(o = babolo_getoptsarg(bos)),             ("No -%c arg", bos->c), EX_USAGE);
                ; PNGTR_O(!!(ex = pngtr_read(cf, o, 0)),                       ("read 0 %u", cf->num));
                ; PNGTR_O(!!(ex = pngtr_realconvey(cf)),                               ("realconvey"));
                ; break;
        case 'I': cf->num = babolo_getoptunum(bos, 0);
                ; PNGTR_W(!!errno,                               ("getoptunum -%c", bos->c), EX_USAGE);
                ; break;
        case 'L': i = babolo_getoptunum(bos, 0);
                ; PNGTR_W((!!errno),                             ("getoptunum -%c", bos->c), EX_USAGE);
                ; PNGTR_X((i < 1) || (i > 7),             ("%d Out of range 1..7", i), EDOM, EX_USAGE);
                ; cf->convey[cf->num].flags |= 0x80 >> i;
                ; break;
        case 'm': PNGTR_W(!(o = babolo_getoptsarg(bos)),             ("No -%c arg", bos->c), EX_USAGE);
                ; PNGTR_X((4 != strlen(o)),           ("invalid -%c %s", bos->c, o), EINVAL, EX_USAGE);
                ; pngtr_crechunk(cf, babolo_testword(&headers, o));
                ; break;
        case 'o': {   pngtr_diapa mimax;

                ;     PNGTR_O(!!(ex = diapa(cf, bos, &mimax)),                ("Err -%c arg", bos->c));
                ;     PNGTR_O(!!(ex = pngtr_opsel(cf, mimax.min, mimax.max)),               ("opsel"));
                ; }
                ; break;
        case 'O': {   pngtr_convey   *convey = &cf->convey[cf->num];

                ;     convey->flags &= ~PNGTR_OPERSEL;
                ;     for (u_int chctr = 0; chctr < convey->chunum; ++chctr) {
                ;         for (u_int opctr = 0; opctr < convey->chunks[chctr].carg; ++opctr) {
                ;             convey->chunks[chctr].operands[opctr].flags &= ~PNGTR_SELECTD;
                ; }   }   }
                ; break;
        case 'p': {   BLIN_flag f = 0;
                ;
                ;     PNGTR_W(!(o = babolo_getoptsarg(bos)),         ("No -%c arg", bos->c), EX_USAGE);
                ;     for (int i = 0; !!o[i]; ++i) {
                ;         switch(o[i]) {
                ;         case 'c': f |= PNGTR_PRINTCD; break;
                          case 'e': f |= PNGTR_PRINTEX; break;
                ;         case 'f': f |= PNGTR_PRINTFM; break;
                ;         case 'F': f |= PNGTR_PRINTFI; break;
                          case 'h': f |= PNGTR_PRINTHD; break;
                ;         case 'i': f |= PNGTR_PRINTIM; break;
                ;         case 'n': f |= PNGTR_PRINTFN; break;
                ;         case 'o': f |= PNGTR_PRINTOP; break;
                ;         case 'p': f |= PNGTR_OUTPAM ; break;
                ;         case 'x': f |= PNGTR_PRINTXD; break;
                ;         default : ifBLIN_QW0("Ill arg -%c %c", bos->c, o[i]);
                ;     }   }
                ;     if  (!cf->file) {
                ;         if  (!cf->ounm) {
                ;             cf->file = stdout;
                ;         } else {
                ;             cf->file = fopen(cf->ounm, "w");
                ;             PNGTR_W(!cf->file,                     ("fopen %s", cf->ounm), EX_IOERR);
                ;         }
                ;         if  (PNGTR_BINOUT & cf->flags) {
                ;             if  (PNGTR_ADDSIGN & cf->flags) {
                ;                 PNGTR_W( (8 != mife_writ(fileno(cf->file), pngtr_sg, 8))
                                         , ("mife_writ")
                                         , EX_IOERR
                                         );
                ;     }   }   }
                ;     if  (PNGTR_BINOUT & cf->flags) {
                ;         PNGTR_O(!!(ex = pngtr_out(cf, f, fileno(cf->file))),                ("out"));
                ;     } else {
                ;         PNGTR_O(!!(ex = pngtr_print(cf, f, cf->file)),                    ("print"));
                ; }   }
                ; break;
        case 'q': BLIN_QUIET(cf->flags);
                ; break;
        case 's': cf->flags |= PNGTR_NOSIGN;
                ; break;
        case 'S': cf->flags |= PNGTR_ADDSIGN;
                ; break;
        case 't': PNGTR_W(!(o = babolo_getoptsarg(bos)),             ("No -%c arg", bos->c), EX_USAGE);
                ; PNGTR_X((4 != strlen(o)),           ("invalid -%c %s", bos->c, o), EINVAL, EX_USAGE);
                ; cf->convey[cf->num].typemask |= (1 << babolo_testword(&headers, o));
                ; cf->convey[cf->num].flags |= PNGTR_CHATYPE;
                ; break;
        case 'T': cf->convey[cf->num].typemask = 0;
                ; cf->convey[cf->num].flags &= ~PNGTR_CHATYPE;
                ; break;
        case 'v': BLIN_VERBOSE(cf->flags);
                ; break;
        case 'x': {   pngtr_diapa mimax;

                ;     PNGTR_O(!!(ex = diapa(cf, bos, &mimax)),                ("Err -%c arg", bos->c));
                ;     PNGTR_O(!!(ex = pngtr_idxsel(cf, mimax.min, mimax.max)),             ("idxsel"));
                ; }
                ; break;
        case 'X': {   pngtr_convey   *convey = &cf->convey[cf->num];

                      convey->flags &= ~PNGTR_INDEXSL;
                ;     for (u_int chctr = 0; chctr < convey->chunum; ++chctr) {
                ;         for (u_int opctr = 0; opctr < convey->chunks[chctr].carg; ++opctr) {
                              if  (!convey->chunks[chctr].operands[opctr].select) {
                                  for ( u_int ixctr = 0
                                      ; ixctr < convey->chunks[chctr].operands[opctr].maxindex
                                      ; ++ixctr) {
                                      convey->chunks[chctr].operands[opctr].select[ixctr] &= ~1;
                ; }   }   }   }   }
                ; break;
        case 'y': cf->flags |= PNGTR_NOINVAL;
                ; break;
        case 'Y': cf->flags |= PNGTR_AGRERR;
                ; break;
        case '=': switch((Tflag = babolo_getoptsarg(bos))[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP)         ; break;
                  case '1': babolo_optest(flar)             ; break;
                  case '2': babolo_optext(bos)              ; break;
                  case '3': babolo_dumpopts(bos)            ; break;
                  case 'T': blin_ctl(BLIN_CTL_TCFN, ++Tflag); break;
                  default : usage(); ex = EX_USAGE          ; break;
                  }
                ; break;
        case '?': usage();
                ; break;
        default : ifBLIN_QX0("Illegal flag -%c", bos->c);
                ; usage();
                ; ex = EX_USAGE;
    }   }
out:
    ifBLIN_QX3("  %d", ex);
    if  (!ex) {
        PNGTR_O(!!(ex = doclo(cf)),                                                         ("doclo"));
    } else {
        PNGTR_O(!!doclo(cf),                                                                ("doclo"));
    }
    if  (cf) {
        for (int i = 0; i < PNGTR_CONVEYNUM; ++i) {
            cf->num = i;
            pngtr_freeconvey(cf);
        }
        free(cf->convey);
        free(cf);
    }
    babolo_closeopts(bos);
    ifBLIN_QX3("- %d", ex);
    exit(ex);
#   undef blin_internal_flags
}
